import logging
import os
from typing import Literal, Union

from rich.console import Console
from rich.logging import RichHandler

scib_logger = logging.getLogger("scib_metrics")


class ScibConfig:
    """Config manager for scib-metrics.

    Examples
    --------
    To set the progress bar style, choose one of "rich", "tqdm"

    >>> scib_metrics.settings.progress_bar_style = "rich"

    To set the verbosity

    >>> import logging
    >>> scib_metrics.settings.verbosity = logging.INFO
    """

    def __init__(
        self,
        verbosity: int = logging.INFO,
        progress_bar_style: Literal["rich", "tqdm"] = "tqdm",
        jax_preallocate_gpu_memory: bool = False,
    ):

        if progress_bar_style not in ["rich", "tqdm"]:
            raise ValueError("Progress bar style must be in ['rich', 'tqdm']")
        self.progress_bar_style = progress_bar_style
        self.jax_preallocate_gpu_memory = jax_preallocate_gpu_memory
        self.verbosity = verbosity

    @property
    def progress_bar_style(self) -> str:
        """Library to use for progress bar."""
        return self._pbar_style

    @progress_bar_style.setter
    def progress_bar_style(self, pbar_style: Literal["tqdm", "rich"]):
        """Library to use for progress bar."""
        self._pbar_style = pbar_style

    @property
    def verbosity(self) -> int:
        """Verbosity level (default `logging.INFO`).

        Returns
        -------
        verbosity: int
        """
        return self._verbosity

    @verbosity.setter
    def verbosity(self, level: Union[str, int]):
        """Set verbosity level.

        If "scib_metrics" logger has no StreamHandler, add one.
        Else, set its level to `level`.

        Parameters
        ----------
        level
            Sets "scib_metrics" logging level to `level`
        force_terminal
            Rich logging option, set to False if piping to file output.
        """
        self._verbosity = level
        scib_logger.setLevel(level)
        if len(scib_logger.handlers) == 0:
            console = Console(force_terminal=True)
            if console.is_jupyter is True:
                console.is_jupyter = False
            ch = RichHandler(level=level, show_path=False, console=console, show_time=False)
            formatter = logging.Formatter("%(message)s")
            ch.setFormatter(formatter)
            scib_logger.addHandler(ch)
        else:
            scib_logger.setLevel(level)

    def reset_logging_handler(self) -> None:
        """Reset "scib_metrics" log handler to a basic RichHandler().

        This is useful if piping outputs to a file.

        Returns
        -------
        None
        """
        scib_logger.removeHandler(scib_logger.handlers[0])
        ch = RichHandler(level=self._verbosity, show_path=False, show_time=False)
        formatter = logging.Formatter("%(message)s")
        ch.setFormatter(formatter)
        scib_logger.addHandler(ch)

    @property
    def jax_preallocate_gpu_memory(self):
        """Jax GPU memory allocation settings.

        If False, Jax will ony preallocate GPU memory it needs.
        If float in (0, 1), Jax will preallocate GPU memory to that
        fraction of the GPU memory.

        Returns
        -------
        jax_preallocate_gpu_memory: bool or float
        """
        return self._jax_gpu

    @jax_preallocate_gpu_memory.setter
    def jax_preallocate_gpu_memory(self, value: Union[float, bool]):
        # see https://jax.readthedocs.io/en/latest/gpu_memory_allocation.html#gpu-memory-allocation
        if value is False:
            os.environ["XLA_PYTHON_CLIENT_PREALLOCATE"] = "false"
        elif isinstance(value, float):
            if value >= 1 or value <= 0:
                raise ValueError("Need to use a value between 0 and 1")
            # format is ".XX"
            os.environ["XLA_PYTHON_CLIENT_MEM_FRACTION"] = str(value)[1:4]
        else:
            raise ValueError("value not understood, need bool or float in (0, 1)")
        self._jax_gpu = value


settings = ScibConfig()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------
import sys
from datetime import datetime
from importlib.metadata import metadata
from pathlib import Path

HERE = Path(__file__).parent
sys.path.insert(0, str(HERE / "extensions"))


# -- Project information -----------------------------------------------------

info = metadata("scib-metrics")
project = info["Name"]
author = info["Author"]
copyright = f"{datetime.now():%Y}, {author}."
version = info["Version"]

# The full version, including alpha/beta/rc tags
release = info["Version"]

bibtex_bibfiles = ["references.bib"]
templates_path = ["_templates"]
nitpicky = True  # Warn about broken links
needs_sphinx = "4.0"

html_context = {
    "display_github": True,  # Integrate GitHub
    "github_user": "yoseflab",  # Username
    "github_repo": project,  # Repo name
    "github_version": "main",  # Version
    "conf_py_path": "/docs/",  # Path in the checkout to the docs root
}

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings.
# They can be extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinxcontrib.bibtex",
    "sphinx_autodoc_typehints",
    "scanpydoc.definition_list_typed_field",
    "scanpydoc.elegant_typehints",
    "nbsphinx",
    "sphinx.ext.mathjax",
    *[p.stem for p in (HERE / "extensions").glob("*.py")],
]

autosummary_generate = True
autodoc_member_order = "groupwise"
default_role = "literal"
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_use_rtype = True  # having a separate entry generally helps readability
napoleon_use_param = True
myst_heading_anchors = 3  # create anchors for h1-h3


intersphinx_mapping = {
    "anndata": ("https://anndata.readthedocs.io/en/stable/", None),
    "ipython": ("https://ipython.readthedocs.io/en/stable/", None),
    "matplotlib": ("https://matplotlib.org/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandas": ("https://pandas.pydata.org/docs/", None),
    "python": ("https://docs.python.org/3", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference/", None),
    "sklearn": ("https://scikit-learn.org/stable/", None),
    "scanpy": ("https://scanpy.readthedocs.io/en/stable/", None),
    "jax": ("https://jax.readthedocs.io/en/latest/", None),
}


nbsphinx_execute = "never"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"
html_static_path = ["_static"]

pygments_style = "default"
pygments_dark_style = "native"

nitpick_ignore = [
    # If building the documentation fails because of a missing link that is outside your control,
    # you can add an exception to this list.
    ("py:class", "np.ndarray"),
    ("py:class", "jnp.ndarray"),
]


def setup(app):
    """App setup hook."""
    app.add_config_value(
        "recommonmark_config",
        {
            "auto_toc_tree_section": "Contents",
            "enable_auto_toc_tree": True,
            "enable_math": True,
            "enable_inline_math": False,
            "enable_eval_rst": True,
        },
        True,
    )



import numpy as np
import pandas as pd

from scib_metrics.utils import silhouette_samples


def silhouette_label(X: np.ndarray, labels: np.ndarray, rescale: bool = True) -> float:
    """Average silhouette width (ASW).

    Parameters
    ----------
    X
        Array of shape (n_samples, n_features).
    labels
        Array of shape (n_samples,) representing label values
    rescale
        Scale asw into the range [0, 1].
    """
    asw = np.mean(silhouette_samples(X, labels))
    if rescale:
        asw = (asw + 1) / 2
    return np.mean(asw)


def silhouette_batch(X: np.ndarray, labels: np.ndarray, batch: np.ndarray, rescale: bool = True) -> float:
    """Average silhouette width (ASW) with respect to batch ids within each label.

    Parameters
    ----------
    X
        Array of shape (n_samples, n_features).
    labels
        Array of shape (n_samples,) representing label values
    batch
        Array of shape (n_samples,) representing batch values
    rescale
        Scale asw into the range [0, 1]. If True, higher values are better.
    """
    sil_dfs = []
    unique_labels = np.unique(labels)
    for group in unique_labels:
        labels_mask = labels == group
        X_subset = X[labels_mask]
        batch_subset = batch[labels_mask]
        n_batches = len(np.unique(batch_subset))

        if (n_batches == 1) or (n_batches == X_subset.shape[0]):
            continue

        sil_per_group = silhouette_samples(X_subset, batch_subset)

        # take only absolute value
        sil_per_group = np.abs(sil_per_group)

        if rescale:
            # scale s.t. highest number is optimal
            sil_per_group = 1 - sil_per_group

        sil_dfs.append(
            pd.DataFrame(
                {
                    "group": [group] * len(sil_per_group),
                    "silhouette_score": sil_per_group,
                }
            )
        )

    sil_df = pd.concat(sil_dfs).reset_index(drop=True)
    sil_means = sil_df.groupby("group").mean()
    asw = sil_means["silhouette_score"].mean()

    return asw



import logging
from importlib.metadata import version

from . import utils
from ._ari_nmi import nmi_ari_cluster_labels
from ._isolated_labels import isolated_labels
from ._settings import settings
from ._silhouette import silhouette_batch, silhouette_label

__all__ = [
    "utils",
    "isolated_labels",
    "silhouette_label",
    "silhouette_batch",
    "nmi_ari_cluster_labels",
]

__version__ = version("scib-metrics")

settings.verbosity = logging.INFO
# Jax sets the root logger, this prevents double output.
logger = logging.getLogger("scib_metrics")
logger.propagate = False



from functools import partial

import jax
import jax.numpy as jnp
import numpy as np
from sklearn.utils import check_array

from ._dist import cdist


def _initialize_random(X: jnp.ndarray, n_clusters: int, key: jnp.ndarray) -> jnp.ndarray:
    """Initialize cluster centroids randomly."""
    n_obs = X.shape[0]
    indices = jax.random.choice(key, n_obs, (n_clusters,), replace=False)
    initial_state = X[indices]
    return initial_state


@jax.jit
def _get_dist_labels(X: jnp.ndarray, centroids: jnp.ndarray) -> jnp.ndarray:
    """Get the distance and labels for each observation."""
    dist = cdist(X, centroids)
    labels = jnp.argmin(dist, axis=1)
    return dist, labels


class KMeansJax:
    """Jax implementation of :class:`sklearn.cluster.KMeans`.

    This implementation is limited to random initialization and euclidean distance.

    Parameters
    ----------
    n_clusters
        Number of clusters.
    n_init
        Number of times the k-means algorithm will be initialized.
    max_iter
        Maximum number of iterations of the k-means algorithm for a single run.
    tol
        Relative tolerance with regards to inertia to declare convergence.
    seed
        Random seed.
    """

    def __init__(self, n_clusters: int = 8, *, n_init: int = 10, max_iter: int = 300, tol: float = 1e-4, seed: int = 0):
        self.n_clusters = n_clusters
        self.n_init = n_init
        self.max_iter = max_iter
        self.tol = tol
        self.seed = seed

    def fit(self, X: np.ndarray):
        """Fit the model to the data."""
        X = check_array(X, dtype=np.float32, order="C")
        self._fit(X)
        return self

    def _fit(self, X: np.ndarray):
        key = jax.random.PRNGKey(self.seed)
        all_centroids, all_inertias = jax.vmap(lambda key: self._kmeans_full_run(X, key))(
            jax.random.split(key, self.n_init)
        )
        i = jnp.argmin(all_inertias)
        self.cluster_centroids_ = np.array(jax.device_get(all_centroids[i]))
        self.inertia_ = np.array(jax.device_get(all_inertias[i]))
        _, labels = _get_dist_labels(X, self.cluster_centroids_)
        self.labels_ = np.array(jax.device_get(labels))

    @partial(jax.jit, static_argnums=(0,))
    def _kmeans_full_run(self, X: jnp.ndarray, key: jnp.ndarray) -> jnp.ndarray:
        def _kmeans_step(state):
            old_inertia = state[1]
            centroids, _, _, n_iter = state
            # TODO(adamgayoso): Efficiently compute argmin and min simultaneously.
            dist, new_labels = _get_dist_labels(X, centroids)
            # From https://colab.research.google.com/drive/1AwS4haUx6swF82w3nXr6QKhajdF8aSvA?usp=sharing
            counts = (new_labels[jnp.newaxis, :] == jnp.arange(self.n_clusters)[:, jnp.newaxis]).sum(
                axis=1, keepdims=True
            )
            counts = jnp.clip(counts, a_min=1, a_max=None)
            # Sum over points in a centroid by zeroing others out
            new_centroids = (
                jnp.sum(
                    jnp.where(
                        # axes: (data points, clusters, data dimension)
                        new_labels[:, jnp.newaxis, jnp.newaxis]
                        == jnp.arange(self.n_clusters)[jnp.newaxis, :, jnp.newaxis],
                        X[:, jnp.newaxis, :],
                        0.0,
                    ),
                    axis=0,
                )
                / counts
            )
            new_inertia = jnp.mean(jnp.min(dist, axis=1))
            n_iter = n_iter + 1
            return new_centroids, new_inertia, old_inertia, n_iter

        def _kmeans_convergence(state):
            _, new_inertia, old_inertia, n_iter = state
            cond1 = jnp.abs(old_inertia - new_inertia) < self.tol
            cond2 = n_iter > self.max_iter
            return jnp.logical_or(cond1, cond2)[0]

        centroids = _initialize_random(X, self.n_clusters, key)
        # centroids, new_inertia, old_inertia, n_iter
        state = (centroids, jnp.inf, jnp.inf, jnp.array([0.0]))
        state = _kmeans_step(state)
        state = jax.lax.while_loop(_kmeans_convergence, _kmeans_step, state)
        return state[0], state[1]



from typing import Tuple

import numpy as np
from sklearn.metrics.cluster import adjusted_rand_score, normalized_mutual_info_score

from .utils import KMeansJax


def _compute_clustering(X: np.ndarray, n_clusters: int) -> np.ndarray:
    kmeans = KMeansJax(n_clusters)
    kmeans.fit(X)
    return kmeans.labels_


def nmi_ari_cluster_labels(X: np.ndarray, labels: np.ndarray) -> Tuple[float, float]:
    """Compute nmi and ari between k-means clusters and labels.

    This deviates from the original implementation in scib by using k-means
    with k equal to the known number of cell types/labels. This leads to
    a more efficient computation of the nmi and ari scores.

    Parameters
    ----------
    X
        Array of shape (n_samples, n_features).
    labels
        Array of shape (n_samples,) representing label values

    Returns
    -------
    nmi: float
        Normalized mutual information score
    ari: float
        Adjusted rand index score
    """
    n_clusters = len(np.unique(labels))
    labels_pred = _compute_clustering(X, n_clusters)
    nmi = normalized_mutual_info_score(labels, labels_pred, average_method="arithmetic")
    ari = adjusted_rand_score(labels, labels_pred)

    return nmi, ari



import jax
import jax.numpy as jnp
import numpy as np


@jax.jit
def _euclidean_distance(x: np.array, y: np.array) -> float:
    dist = jnp.sqrt(jnp.sum((x - y) ** 2))
    return dist


@jax.jit
def cdist(x: np.ndarray, y: np.ndarray) -> jnp.ndarray:
    """Jax implementation of :func:`scipy.spatial.distance.cdist`.

    Uses euclidean distance.

    Parameters
    ----------
    x
        Array of shape (n_samples_a, n_features)
    y
        Array of shape (n_samples_b, n_features)

    Returns
    -------
    dist: jnp.ndarray
        Array of shape (n_samples_a, n_samples_b)
    """
    return jax.vmap(lambda x1: jax.vmap(lambda y1: _euclidean_distance(x1, y1))(y))(x)



import jax.numpy as jnp
import numpy as np
from scipy.spatial.distance import cdist as sp_cdist
from sklearn.metrics import silhouette_samples as sk_silhouette_samples

import scib_metrics


def dummy_x_labels():
    X = np.array([[1, 2], [3, 4], [5, 6], [7, 8], [9, 10], [11, 12]])
    labels = np.array([0, 0, 1, 1, 0, 1])
    return X, labels


def dummy_x_labels_batch():
    X, labels = dummy_x_labels()
    batch = np.array([0, 1, 0, 1, 0, 1])
    return X, labels, batch


def test_package_has_version():
    scib_metrics.__version__


def test_cdist():
    x = jnp.array([[1, 2], [3, 4]])
    y = jnp.array([[5, 6], [7, 8]])
    assert np.allclose(scib_metrics.utils.cdist(x, y), sp_cdist(x, y))


def test_silhouette_samples():
    X, labels = dummy_x_labels()
    assert np.allclose(scib_metrics.utils.silhouette_samples(X, labels), sk_silhouette_samples(X, labels))


def test_silhouette_label():
    X, labels = dummy_x_labels()
    score = scib_metrics.silhouette_label(X, labels)
    assert score > 0
    scib_metrics.silhouette_label(X, labels, rescale=False)


def test_silhouette_batch():
    X, labels, batch = dummy_x_labels_batch()
    score = scib_metrics.silhouette_batch(X, labels, batch)
    assert score > 0
    scib_metrics.silhouette_batch(X, labels, batch)


def test_nmi_ari_cluster_labels():
    X, labels = dummy_x_labels()
    nmi, ari = scib_metrics.nmi_ari_cluster_labels(X, labels)
    assert isinstance(nmi, float)
    assert isinstance(ari, float)


def test_isolated_labels():
    X, labels, batch = dummy_x_labels_batch()
    scib_metrics.isolated_labels(X, labels, batch)


def test_kmeans():
    X, _ = dummy_x_labels()
    kmeans = scib_metrics.utils.KMeansJax(2)
    kmeans.fit(X)
    assert kmeans.labels_.shape == (X.shape[0],)



import logging
from typing import Optional, Union

import numpy as np
import pandas as pd

from ._silhouette import silhouette_label

logger = logging.getLogger(__name__)


def isolated_labels(
    X: np.ndarray,
    labels: np.ndarray,
    batch: np.ndarray,
    iso_threshold: Optional[int] = None,
) -> float:
    """Isolated label score.

    Score how well labels of isolated labels are distiguished in the dataset by
    average-width silhouette score (ASW) on isolated label vs all other labels.

    The default of the original scib package is to use a cluster-based F1 scoring
    procedure, but here we use the ASW for speed and simplicity.

    Parameters
    ----------
    X
        Array of shape (n_samples, n_features).
    labels
        Array of shape (n_samples,) representing label values
    batch
        Array of shape (n_samples,) representing batch values
    iso_threshold
        Max number of batches per label for label to be considered as
        isolated, if integer. If `None`, considers minimum number of
        batches that labels are present in

    Returns
    -------
    isolated_label_score: float
    """
    scores = {}
    isolated_labels = _get_isolated_labels(labels, batch, iso_threshold)

    for label in isolated_labels:
        score = _score_isolated_label(X, labels, label)
        scores[label] = score
    scores = pd.Series(scores)

    return scores.mean()


def _score_isolated_label(
    X: np.ndarray,
    labels: np.ndarray,
    isolated_label: Union[str, float, int],
):
    """Compute label score for a single label."""
    mask = labels == isolated_label
    score = silhouette_label(X, mask.astype(np.float32))
    logging.info(f"{isolated_label}: {score}")

    return score


def _get_isolated_labels(labels: np.ndarray, batch: np.ndarray, iso_threshold: float):
    """Get labels that are isolated depending on the number of batches."""
    tmp = pd.DataFrame()
    label_key = "label"
    batch_key = "batch"
    tmp[label_key] = labels
    tmp[batch_key] = batch
    tmp = tmp.drop_duplicates()
    batch_per_lab = tmp.groupby(label_key).agg({batch_key: "count"})

    # threshold for determining when label is considered isolated
    if iso_threshold is None:
        iso_threshold = batch_per_lab.min().tolist()[0]

    logging.info(f"isolated labels: no more than {iso_threshold} batches per label")

    labels = batch_per_lab[batch_per_lab[batch_key] <= iso_threshold].index.tolist()
    if len(labels) == 0:
        logging.info(f"no isolated labels with less than {iso_threshold} batches")

    return np.array(labels)



import itertools
from dataclasses import dataclass
from typing import Union

import jax
import jax.numpy as jnp
import numpy as np

from ._dist import cdist

NdArray = Union[np.ndarray, jnp.ndarray]


@dataclass
class _IntraClusterData:
    data: NdArray
    labels: NdArray
    unique_labels: NdArray
    intra_dist: NdArray


@dataclass
class _InterClusterData:
    data: NdArray
    labels: NdArray
    inter_dist: NdArray
    label_combos: NdArray


def _intra_cluster_distances(X: np.ndarray, labels: np.ndarray):
    """Calculate the mean intra-cluster distance."""
    intra_dist = jnp.zeros((X.shape[0],))
    unique_labels = jnp.unique(labels)
    # TODO(adamgayoso) See if we can lax this
    # output = jax.lax.fori_loop(0, len(unique_labels), _intra_cluster_distances_block, input)
    for i in range(len(unique_labels)):
        input = _IntraClusterData(X, labels, unique_labels, intra_dist)
        intra_dist = _intra_cluster_distances_block(i, input)
    return intra_dist


@jax.jit
def _intra_value(subset: np.ndarray) -> jnp.ndarray:
    distances = cdist(subset, subset)
    values = distances.sum(axis=1) / (distances.shape[0] - 1)
    return values


def _intra_cluster_distances_block(i: int, input: _IntraClusterData) -> jnp.ndarray:
    labels_inds = input.labels == input.unique_labels[i]
    subset = input.data[labels_inds]
    values = _intra_value(subset)
    intra_dist = input.intra_dist
    intra_dist = intra_dist.at[labels_inds].set(values)
    return intra_dist


def _nearest_cluster_distances(X: np.ndarray, labels: np.ndarray):
    """Calculate the mean nearest-cluster distance for observation i."""
    unique_labels = jnp.unique(labels)
    inter_dist = jnp.array(np.inf * np.ones((X.shape[0],)))
    label_combinations = jnp.array([list(i) for i in list(itertools.combinations(unique_labels, 2))])
    for i in range(len(label_combinations)):
        input = _InterClusterData(X, labels, inter_dist, label_combinations)
        inter_dist = _nearest_cluster_distance_block(i, input)
    return inter_dist


@jax.jit
def _inter_values(subset_a: np.ndarray, subset_b: np.ndarray) -> Union[jnp.ndarray, jnp.ndarray]:
    distances = cdist(subset_a, subset_b)
    values_a = distances.mean(axis=1)
    values_b = distances.mean(axis=0)
    return values_a, values_b


def _nearest_cluster_distance_block(inter_dist: jnp.ndarray, input: _InterClusterData) -> jnp.ndarray:
    label_a = input.label_combos[inter_dist, 0]
    label_b = input.label_combos[inter_dist, 1]
    label_mask_a = input.labels == label_a
    label_mask_b = input.labels == label_b
    subset_a = input.data[label_mask_a]
    subset_b = input.data[label_mask_b]
    dist_a, dist_b = _inter_values(subset_a, subset_b)
    inter_dist = input.inter_dist
    inter_dist = inter_dist.at[label_mask_a].set(jnp.minimum(dist_a, inter_dist[label_mask_a]))
    inter_dist = inter_dist.at[label_mask_b].set(jnp.minimum(dist_b, inter_dist[label_mask_b]))
    return inter_dist


def silhouette_samples(X: np.ndarray, labels: np.ndarray) -> np.ndarray:
    """Compute the Silhouette Coefficient for each observation.

    Code inspired by:
    https://github.com/maxschelski/pytorch-cluster-metrics/

    Parameters
    ----------
    X
        Array of shape (n_samples, n_features) representing a
        feature array.
    labels
        Array of shape (n_samples,) representing label values
        for each observation.

    Returns
    -------
    silhouette: np.ndarray
        Array of shape (n_samples,)
        Silhouette Coefficients for each sample.
    """
    if X.shape[0] != labels.shape[0]:
        raise ValueError("X and labels should have the same number of samples")

    intra_dist = _intra_cluster_distances(X, labels)
    inter_dist = _nearest_cluster_distances(X, labels)
    return jax.device_get((inter_dist - intra_dist) / jnp.maximum(intra_dist, inter_dist))
